# Kdump anaconda configuration
#
# Copyright (C) 2013 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Shea <dshea@redhat.com>
#

"""Kdump anaconda TUI configuration"""

import os.path
import re

from pyanaconda.modules.common.constants.services import STORAGE
from pyanaconda.core.kernel import kernel_arguments
from pyanaconda.ui.categories.system import SystemCategory
from pyanaconda.ui.tui.spokes import NormalTUISpoke
from pyanaconda.ui.tui.tuiobject import Dialog
from simpleline.render.widgets import CheckboxWidget, EntryWidget, TextWidget
from simpleline.render.containers import ListColumnContainer
from simpleline.render.screen import InputState
from com_redhat_kdump.common import getMemoryBounds, getLuksDevices
from com_redhat_kdump.i18n import N_, _
from com_redhat_kdump.constants import FADUMP_CAPABLE_FILE, ENCRYPTION_WARNING

__all__ = ["KdumpSpoke"]

class KdumpSpoke(NormalTUISpoke):
    category = SystemCategory
    help_id = "KdumpSpoke"

    def __init__(self, *args):
        super().__init__(*args)
        self.title = N_("Kdump")
        self._addon_data = self.data.addons.com_redhat_kdump

        self._lower, self._upper, self._step = getMemoryBounds()
        # Allow either "auto" or a string of digits optionally followed by 'M'
        self._reserve_check_re = re.compile(r'^((auto)|(\d+M?))$')

        self._container = None

        self._luks_devs = []
        self._ready = True

    def _check_storage_change(self, interface, changed, invalid):
        self._ready = False
        partition = changed.get("AppliedPartitioning")
        if partition:
            self._luks_devs = getLuksDevices(partition.unpack())
        self._ready = True

    @property
    def ready(self):
        return self._ready

    def initialize(self):
        # Connect a callback to the PropertiesChanged signal.
        storage = STORAGE.get_proxy()
        storage.PropertiesChanged.connect(self._check_storage_change)

    @classmethod
    def should_run(cls, environment, data):
        # the KdumpSpoke should run only if requested
        return kernel_arguments.get("kdump_addon", default=True) not in ["0", "off", False]

    def apply(self):
        pass

    @property
    def completed(self):
        return True

    @property
    def status(self):
        if not self.data.addons.com_redhat_kdump.enabled:
            return _("Kdump is disabled")
        if not self._ready:
            return _("Checking storage...")
        if self._luks_devs:
            return _("Kdump may require extra setup for encrypted devices.")
        return _("Kdump is enabled")

    def refresh(self, args=None):
        super().refresh(args)

        self._container = ListColumnContainer(1)
        self.window.add(self._container)

        self._create_enable_checkbox()

        if self._addon_data.enabled:
            self._create_fadump_checkbox()
            self._create_reserve_amount_text_widget()

        if self._luks_devs:
            self.window.add_separator()
            message = TextWidget(_(ENCRYPTION_WARNING))
            self.window.add(message)

        self.window.add_separator()

    def _create_enable_checkbox(self):
        enable_kdump_checkbox = CheckboxWidget(title=_("Enable kdump"),
                                               completed=self._addon_data.enabled)
        self._container.add(enable_kdump_checkbox, self._set_enabled)

    def _create_fadump_checkbox(self):
        if not os.path.exists(FADUMP_CAPABLE_FILE):
            return

        enable_fadump_checkbox = CheckboxWidget(title=_("Enable dump mode fadump"),
                                                completed=self._addon_data.enablefadump)
        self._container.add(enable_fadump_checkbox, self._set_fadump_enable)

    def _create_reserve_amount_text_widget(self):
        title = _("Reserve amount (%d - %d MB)" % (self._lower, self._upper))
        reserve_amount_entry = EntryWidget(title=title, value=self._addon_data.reserveMB)
        self._container.add(reserve_amount_entry, self._get_reserve_amount)

    def _set_enabled(self, data):
        self._addon_data.enabled = not self._addon_data.enabled

    def _set_fadump_enable(self, data):
        self._addon_data.enablefadump = not self._addon_data.enablefadump

    def _get_reserve_amount(self, data):
        text = "Reserve amount (%d - %d MB)" % (self._lower, self._upper)
        dialog = Dialog(title=text, conditions=[self._check_reserve_valid])

        self._addon_data.reserveMB = dialog.run()

    def _check_reserve_valid(self, key, report_func):
        if self._reserve_check_re.match(key):
            if key == "auto":
                return True
            if key[-1] == 'M':
                key = key[:-1]
            key = int(key)
            if self._upper >= key >= self._lower:
                return True
        return False

    def input(self, args, key):
        if self._container.process_user_input(key):
            self.redraw()
            return InputState.PROCESSED
        else:
            return super().input(args, key)



# Kdump configuration common methods
#
# Copyright (C) 2014 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Shea <dshea@redhat.com>
#
import re
__all__ = ["getReservedMemory", "getTotalMemory", "getMemoryBounds", "getLuksDevices"]

import blivet.arch
from pyanaconda.modules.common.structures.storage import DeviceData
from pyanaconda.modules.common.constants.services import STORAGE

_reservedMemory = None
def getReservedMemory():
    """Return the amount of memory currently reserved for kdump in MB."""
    global _reservedMemory

    # Check if the value has already been read
    if _reservedMemory is not None:
        return _reservedMemory

    try:
        with open("/sys/kernel/kexec_crash_size", "r") as fobj:
            _reservedMemory = int(fobj.read()) / (1024*1024)
            return _reservedMemory
    except (ValueError, IOError):
        return 0

def getTotalMemory():
    """Return the total amount of system memory in MB

       This is the amount reported by /proc/meminfo plus the aount
       currently reserved for kdump.
    """
    memkb = 0
    fd = open('/proc/meminfo').read()
    matched = re.search(r'^MemTotal:\s+(\d+)', fd)
    if matched:
        memkb = int(matched.groups()[0])

    # total_memory return memory in KB, convert to MB
    availMem = memkb / 1024

    return availMem + getReservedMemory()

def getMemoryBounds():
    """Return a tuple of (lower, upper, step) for kdump reservation limits.

       If there is not enough memory available to use kdump, both lower and
       upper will be 0.
    """

    totalMemory = getTotalMemory()
    arch = blivet.arch.get_arch()

    if arch.startswith('ppc64'):
        lowerBound = 384
        minUsable = 1024
        step = 1
    elif arch == 'aarch64':
        lowerBound = 512
        minUsable = 512
        step = 1
    else:
        lowerBound = 160
        minUsable = 512
        step = 1

    upperBound = (totalMemory - minUsable) - (totalMemory % step)

    if upperBound < lowerBound:
        upperBound = lowerBound = 0

    return (lowerBound, upperBound, step)

def getLuksDevices(object_path):
    devs = []

    partitioning = STORAGE.get_proxy(object_path)
    device_tree = STORAGE.get_proxy(partitioning.GetDeviceTree())
    devices = device_tree.GetDevices()

    for device_name in devices:
        device_data = DeviceData.from_structure(
            device_tree.GetDeviceData(device_name)
        )

        if device_data.type == 'luks/dm-crypt':
            devs.append(device_name)

    return devs



from utils import KdumpTestCase
from unittest.mock import patch
from com_redhat_kdump.ks.kdump import KdumpData
from pykickstart.errors import KickstartParseError
import re


ALL_ARGS_RE = re.compile(
    r"[\s\n]*%addon\s+KdumpData\s+"
    r"(--enable|--disable|--enablefadump|--reverse-mb[=|\ ]['\"]?\d+['\"][Mm]?)*"
    r"\s+.*[\s\n]+%end[\s\n]*")

def new_ks_addon_data():
    return KdumpData("KdumpData")


def kdump_check_ks(test, addon_data, required_arguments):
    ks_str = str(addon_data)
    for arg in required_arguments:
        # Ensure rerquired argument is generated.
        arg_regex = re.compile(r"[\s\n]*%addon\s+KdumpData.*\s+" + arg + r"\s+.*[\s\n]+%end[\s\n]*")
        test.assertRegexpMatches(ks_str, arg_regex)
    # Ensure no invalid argument is generated.
    test.assertRegexpMatches(ks_str, ALL_ARGS_RE)


class KdumpKickstartTestCase(KdumpTestCase):
    def new_ks_addon_data_test(self):
        ks_addon_data = new_ks_addon_data()
        self.assertIsNotNone(ks_addon_data)

    def ks_default_to_str_test(self):
        ks_addon_data = new_ks_addon_data()
        kdump_check_ks(self, ks_addon_data, ["--enable"])

    def ks_enable_to_str_test(self):
        ks_addon_data = new_ks_addon_data()
        ks_addon_data.enabled = True
        kdump_check_ks(self, ks_addon_data, ["--enable"])

    def ks_disable_to_str_test(self):
        ks_addon_data = new_ks_addon_data()
        ks_addon_data.enabled = False
        kdump_check_ks(self, ks_addon_data, ["--disable"])

    def ks_parse_enable_to_str_test(self):
        ks_addon_data = new_ks_addon_data()
        ks_addon_data.handle_header(0, ["--enable"])
        kdump_check_ks(self, ks_addon_data, ["--enable"])

    def ks_parse_reserve_to_str_test(self):
        ks_addon_data = new_ks_addon_data()
        ks_addon_data.handle_header(0, ["--enable", "--reserve-mb=256"])
        kdump_check_ks(self, ks_addon_data, ["--enable", "--reserve-mb[=|\ ](\"256\"|'256'|256)[Mm]?"])

    def ks_parse_fadump_to_str_test(self):
        ks_addon_data = new_ks_addon_data()
        ks_addon_data.handle_header(0, ['--enablefadump'])
        kdump_check_ks(self, ks_addon_data, ["--enablefadump"])

    def ks_parse_invalid_reserve_size_test(self):
        with self.assertRaises(KickstartParseError) as error:
            ks_addon_data = new_ks_addon_data()
            ks_addon_data.handle_header(0, ['--reserve-mb='])
        with self.assertRaises(KickstartParseError) as error:
            ks_addon_data = new_ks_addon_data()
            ks_addon_data.handle_header(0, ['--reserve-mb=invalid'])



import unittest

from unittest.mock import patch
from com_redhat_kdump import common

def enable_kdump_addon_in_anaconda():
    return patch('pyanaconda.kernel.kernel_arguments.is_enabled', return_value=True)

class KdumpTestCase(unittest.TestCase):
    def setUp(self):
        # Clean up global variable that may cache test result of previous test case
        common._reservedMemory = None



from utils import KdumpTestCase
from unittest.mock import patch, mock_open, MagicMock
from com_redhat_kdump import common

SYS_CRASH_SIZE = '/sys/kernel/kexec_crash_size'
PROC_MEMINFO = '/proc/meminfo'

X86_INFO_FIXTURE = {
    SYS_CRASH_SIZE: "167772160",  # 160MB
    PROC_MEMINFO:"""MemTotal:        4030464 kB
"""  # 4GB - 160MB
}

AARCH64_INFO_FIXTURE = {
    SYS_CRASH_SIZE: "536870912",  # 512MB
    PROC_MEMINFO:"""MemTotal:       66584576 kB
"""  # 64GB - 512MB
}

PPC64_INFO_FIXTURE = {
    SYS_CRASH_SIZE: "1073741824",  # 1024MB
    PROC_MEMINFO:"""MemTotal:       66060288 kB
"""  # 64GB - 1GB
}


class MockFileRead(MagicMock):
    def __init__(self, file_read_map):
        MagicMock.__init__(self, name=open, spec=open)
        self.file_read_map = file_read_map

        handle = MagicMock()
        handle.__enter__.return_value = handle
        handle.read.return_value = None

        def reset_choose_file(filename, *args, **kwargs):
            handle.read.return_value = self.file_read_map[filename]
            return handle

        self.side_effect = reset_choose_file


class KdumpCommonTestCase(KdumpTestCase):
    @patch("builtins.open", MockFileRead(X86_INFO_FIXTURE))
    @patch("blivet.arch.get_arch", return_value="x86_64")
    def memory_bound_test_x86(self, _mock_read):
        self.assertEqual((160, 4 * 1024 - 512, 1), common.getMemoryBounds())

    @patch("builtins.open", MockFileRead(AARCH64_INFO_FIXTURE))
    @patch("blivet.arch.get_arch", return_value="aarch64")
    def memory_bound_test_aarch64(self, _mock_read):
        self.assertEqual((512, 64 * 1024 - 512, 1), common.getMemoryBounds())

    @patch("builtins.open", MockFileRead(PPC64_INFO_FIXTURE))
    @patch("blivet.arch.get_arch", return_value="ppc64")
    def memory_bound_test_ppc64(self, _mock_read):
        self.assertEqual((384, 64 * 1024 - 1024, 1), common.getMemoryBounds())
